//
// ringbufferpuller.h
// Copyright Phase Five Systems
//
// Pulls audio into a local ring buffer
//

#include "jump/audio/ringbufferpuller.h"
#include "jump/audio/wavwriter.h"

using namespace jump;

static const TimeVal s_DefaultPullInterval = TimeVal::FromMilliseconds(10);

AudioRingBufferPublisher::AudioRingBufferPublisher(
    Logger::LogCompRef parentLogger)
    : mLogger(parentLogger->CreateComponent("audio.ring.pub"))
    , mPullerThread(ThreadRunnerStdThread::
                        CreateRunnerThread("com.p5sys.audio.ring.pub.puller",
                                           ThreadPriority::Highest))
    , mPollInterval(s_DefaultPullInterval)
    , mRunner(mPullerThread.get())
    , mConns("adpul", mPullerThread.get())
{
    mAudioDesc.format = RDC_AUDIO_FORMAT_PCM;
    mAudioDesc.bitsPerSample = 16;
    mAudioDesc.numChannels = 2;
    mAudioDesc.sampleRateHZ = 48000;

    mAudioPullPoller.SetPoller(mPullerThread.get(),
                               mPollInterval,
                               [=]() { PullAudio(); });
}

AudioRingBufferPublisher::~AudioRingBufferPublisher()
{
    Stop();
}

void AudioRingBufferPublisher::SetPuller(
    std::shared_ptr<RDClientAudioStreamPuller> puller)
{
    mRunner.RunAsync(
        [=]()
        {
            mConns.disconnect_all();
            mPuller = puller;

            if (mPuller)
            {
                mPuller->AudioAvailable
                    .ObserveAsync(mConns,
                                  [=](bool) { mAudioPullPoller.Start(); });
            }

            mAudioPullPoller.Start();
        });
}

void AudioRingBufferPublisher::SetRingBuffer(
    std::shared_ptr<LocalAudioCaptureMappedRingBuffer> buffer,
    std::shared_ptr<SharedMem> optionalBufferBackingSharedMem)
{
    mRunner.RunAsync(
        [=]()
        {
            mRingBuffer = buffer;
            mRingBufferSharedMem = optionalBufferBackingSharedMem;
            mAudioPullPoller.Start();
        });
}

void AudioRingBufferPublisher::SetDebugOutputFilePath(
    const std::string &outputFilePath)
{
    mDebugOutputFilePath = outputFilePath;
}

void AudioRingBufferPublisher::PullAudio()
{
    if (!mPuller || !mRingBuffer || !mPuller->AudioAvailable)
    {
        LOG_TRACE(mLogger, "Stopping puller");
        mAudioPullPoller.Stop();
        return;
    }

    uint8_t buffer[1024 * 256];
    auto bytesRequired = mAudioDesc.GetBytesRequiredForTime(mPollInterval);
    if (bytesRequired >= sizeof(buffer))
    {
        DASSERT(false);
        LOG_ERROR(mLogger, "Buffer is too small");
        return;
    }

    LOG_TRACE(mLogger, "Pulling audio: {}ms", mPollInterval.ToMilliseconds());

    auto framesToPull = bytesRequired / mAudioDesc.GetFrameSize();

    // Read
    mPuller->SignalGetAudio.fire(mAudioDesc, framesToPull, buffer);

    // Write to the debug file if needed
    if (mDebugOutputFilePath.length() > 0)
    {
        WaveWriter::Append(mDebugOutputFilePath,
                           mAudioDesc,
                           buffer,
                           bytesRequired);
    }

    // Write
    mRingBuffer->Write(buffer, bytesRequired);
    mRingBuffer->CommitWrite(bytesRequired);
}

void AudioRingBufferPublisher::Start()
{
    mRunner.RunAsync(
        [=]()
        {
            LOG_TRACE(mLogger, "Starting audio");
            mAudioPullPoller.Start();
        });
}

void AudioRingBufferPublisher::Stop()
{
    mRunner.RunAsync(
        [=]()
        {
            LOG_TRACE(mLogger, "Stopping audio!");
            mAudioPullPoller.Stop();
        });
}

// ------------------------------------------------------------------------
AudioRingBufferConsumer::AudioRingBufferConsumer(
    Logger::LogCompRef parentLogger)
    : mLogger(parentLogger->CreateComponent("audio.ring.consume"))
{
    // nothing
}

void AudioRingBufferConsumer::ReduceLagIfNeeded(
    const RDClientAudioStreamDescription &desc,
    LocalAudioCaptureMappedRingBuffer &buffer,
    const TimeVal &audioProducerPushInterval,
    const TimeVal &audioReaderPullInterval)
{
    // Maximum lag is whatever is greater producer or reader lag * 2
    const auto minProducerLagTime = audioProducerPushInterval * 2;
    const auto minReaderLagTime = audioReaderPullInterval * 2;
    const auto maxLagTime = minProducerLagTime > minReaderLagTime
                                ? minProducerLagTime
                                : minReaderLagTime;

    const auto reduceLagIncrement =
        desc.GetBytesRequiredForTime(TimeVal::FromMilliseconds(1));
    const auto minLagBytesRequired = desc.GetBytesRequiredForTime(maxLagTime);

    // Make sure we're not lagging too far behind the audio
    while (buffer.Size() > minLagBytesRequired)
    {
        LOG_DEBUG(mLogger,
                  "Reducing lag - max lag = {}ms, current lag = {}ms",
                  maxLagTime.ToMilliseconds(),
                  desc.GetTimeFromSampleSize(buffer.Size()).ToMilliseconds());
        buffer.IncrementReadPointer(reduceLagIncrement);
    }
}

void AudioRingBufferConsumer::OnGetAudio(
    const RDClientAudioStreamDescription &desc,
    int frames,
    void *destBuffer,
    LocalAudioCaptureMappedRingBuffer &buffer)
{
    // Reduce the lag in the ring buffer
    ReduceLagIfNeeded(desc,
                      buffer,
                      s_DefaultPullInterval,
                      desc.GetTimeFromFrames(frames));

    // Read in the frame buffer
    int bytesToRead = frames * desc.GetFrameSize();
    auto ringBufferSize = buffer.Size();
    int silenceBytes = 0;

    LOG_TRACE(mLogger, "Ringbuffer size is: {}", ringBufferSize);

    if (bytesToRead > ringBufferSize)
    {
        LOG_DEBUG(mLogger,
                  "Don't have enough audio: Audio input wants to read {}, "
                  "Buffer size is: {}",
                  bytesToRead,
                  ringBufferSize);
        silenceBytes = bytesToRead - ringBufferSize;
        bytesToRead = ringBufferSize;
    }

    uint8_t *destBuffer8 = reinterpret_cast<uint8_t *>(destBuffer);
    buffer.Read(destBuffer8, bytesToRead);
    buffer.IncrementReadPointer(bytesToRead);

    // Add silience if needed
    memset(&destBuffer8[bytesToRead], 0, silenceBytes);
}
